/*
    PyGame - Python Game Library
    Copyright (C) 2000  Pete Shinners

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Pete Shinners
    pete@shinnerso.org
*/

#define PYGAMEAPI_CDROM_INTERNAL
#include "pygame.h"




staticforward PyTypeObject PyCD_Type;
static PyObject* PyCD_New(SDL_CD* cdrom);
#define PyCD_Check(x) ((x)->ob_type == &PyCD_Type)



static void cdrom_autoquit()
{
	if(SDL_WasInit(SDL_INIT_CDROM))
		SDL_QuitSubSystem(SDL_INIT_CDROM);
}

static PyObject* cdrom_autoinit(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	if(!SDL_WasInit(SDL_INIT_CDROM))
	{
		if(SDL_InitSubSystem(SDL_INIT_CDROM))
			return PyInt_FromLong(0);
		PyGame_RegisterQuit(cdrom_autoquit);
	}
	return PyInt_FromLong(1);
}


static char cdrom_quit_doc[] =
	"quit() -> None\n"
	"uninitialize the cdrom subsystem";

static PyObject* cdrom_quit(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	cdrom_autoquit();

	RETURN_NONE
}


static char cdrom_init_doc[] =
	"init() -> None\n"
	"initialize the cdrom subsystem";

static PyObject* cdrom_init(PyObject* self, PyObject* arg)
{
	PyObject* result;
	int istrue;

	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	result = cdrom_autoinit(self, arg);
	istrue = PyObject_IsTrue(result);
	Py_DECREF(result);
	if(!istrue)
		return RAISE(PyExc_SDLError, SDL_GetError());

	RETURN_NONE
}



static char get_init_doc[] = 
	"get_init() -> bool\n"
	"returns true if cdrom subsytem is initialized";

static PyObject* get_init(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	return PyInt_FromLong(SDL_WasInit(SDL_INIT_CDROM)!=0);
}


static void cd_dealloc(PyObject* self)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(SDL_WasInit(SDL_INIT_CDROM))
		SDL_CDClose(cd_ref->cd);
	PyMem_DEL(self);	
}



static char cdrom_open_doc[] =
	"open(id) -> CD\n"
	"retrieve a CD object for the given id";

static PyObject* cdrom_open(PyObject* self, PyObject* args)
{
	int id;
	
	if(!PyArg_ParseTuple(args, "i", &id))
		return NULL;

	CDROM_INIT_CHECK();

	return PyCD_New(SDL_CDOpen(id));
}



static char cdrom_count_doc[] =
	"count() -> int\n"
	"returns the number of CD-ROM drives installed";

static PyObject* cdrom_count(PyObject* self, PyObject* args)
{
	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	CDROM_INIT_CHECK();

	return PyInt_FromLong(SDL_CDNumDrives());
}



static char cdrom_name_doc[] =
	"name(id) -> string\n"
	"return the name of the CD-ROM drive whose id is specified";

static PyObject* cdrom_name(PyObject* self, PyObject* args)
{
	int drive;
	
	if(!PyArg_ParseTuple(args, "i", &drive))
		return NULL;

	CDROM_INIT_CHECK();

	return PyString_FromString(SDL_CDName(drive));
}



static char cd_play_tracks_doc[] =
	"play_tracks(start_track, start_frame, ntracks, nframes) -> int\n"
	"Start playing from start_track, for ntracks and nframes.\n"
	"If ntracks and nframes are 0, it will play until the end of the CD";

static PyObject* cd_play_tracks(PyObject* self, PyObject* args)
{
	int start_track, start_frame, ntracks, nframes;
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, "iiii", &start_track, &start_frame, &ntracks,
		&nframes	))
		return NULL;
		
	SDL_CDStatus(cd_ref->cd);
		
	return PyInt_FromLong(
		SDL_CDPlayTracks(cd_ref->cd, start_track, start_frame, ntracks, nframes));
}



static char cd_play_doc[] =
	"play(start_frame, nframes) -> int\n"
	"Start playing from start_frame for nframes.";

static PyObject* cd_play(PyObject* self, PyObject* args)
{
	int start_frame, nframes;
	PyCDObject* cd_ref = (PyCDObject*)self;
	
	if(!PyArg_ParseTuple(args, "ii", &start_frame, &nframes))
		return NULL;

	SDL_CDStatus(cd_ref->cd);

	return PyInt_FromLong(SDL_CDPlay(cd_ref->cd, start_frame, nframes));
}



static char cd_pause_doc[] =
	"pause() -> int\n"
	"Pause the currently playing CD.";

static PyObject* cd_pause(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(cd_ref->cd) != CD_PLAYING)
		return PyInt_FromLong(-1);

	return PyInt_FromLong(SDL_CDPause(cd_ref->cd));
}



static char cd_resume_doc[] =
	"resume() -> int\n"
	"Resume the currently paused CD.";

static PyObject* cd_resume(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(cd_ref->cd) != CD_PAUSED)
		return PyInt_FromLong(-1);

	return PyInt_FromLong(SDL_CDResume(cd_ref->cd));
}



static char cd_stop_doc[] =
	"stop() -> int\n"
	"Stop the CD.";

static PyObject* cd_stop(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(cd_ref->cd) < CD_PLAYING)
		return PyInt_FromLong(-1);

	return PyInt_FromLong(SDL_CDStop(cd_ref->cd));
}



static char cd_eject_doc[] =
	"eject() -> int\n"
	"Eject the CD.";

static PyObject* cd_eject(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(cd_ref->cd) <= 0 )
		return PyInt_FromLong(-1);

	return PyInt_FromLong(SDL_CDEject(cd_ref->cd));
}



static char cd_status_doc[] =
	"get_status() -> int\n"
	"Returns the status of the CD object.";

static PyObject* cd_status(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_CDStatus(cd_ref->cd));
}



static char cd_cur_track_doc[] =
	"get_cur_track() -> int\n"
	"Returns the current track's position, or -1.";

static PyObject* cd_cur_track(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;
	SDL_CD* s_cd_ref = cd_ref->cd;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(s_cd_ref) <= 1)
		return PyInt_FromLong(-1);

	return PyInt_FromLong(s_cd_ref->cur_track);
}



static char cd_cur_frame_doc[] =
	"get_cur_frame() -> int\n"
	"Returns the current frame position, or -1.";

static PyObject* cd_cur_frame(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;
	SDL_CD* s_cd_ref = cd_ref->cd;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if(SDL_CDStatus(s_cd_ref) <= 1)
		return PyInt_FromLong(-1);

	return PyInt_FromLong(s_cd_ref->cur_frame);
}



static char cd_get_empty_doc[] =
	"get_empty() -> int\n"
	"Returns 0 if there is no CD in the drive, or 1 otherwise.";

static PyObject* cd_get_empty(PyObject* self, PyObject* args)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(CD_INDRIVE(SDL_CDStatus(cd_ref->cd)));
}





static int cd_seq_len(PyObject* self)
{
	PyCDObject* cd_ref = (PyCDObject*)self;

	if(!SDL_WasInit(SDL_INIT_CDROM))
	{
		RAISE(PyExc_RuntimeError, "SDL has shut down.");
		return -1;
	}
	
	if(!CD_INDRIVE(SDL_CDStatus(cd_ref->cd)))
	{
		RAISE(PyExc_RuntimeError, "No CD present");
		return -1;
	}

	return cd_ref->cd->numtracks;
}



static PyObject* cd_seq_get(PyObject* self, int index)
{
	PyCDObject* cd_ref = (PyCDObject*)self;
	SDL_CD* s_cd_ref = cd_ref->cd;
	SDL_CDtrack* s_track_ref;

	if(!SDL_WasInit(SDL_INIT_CDROM))
		return RAISE(PyExc_RuntimeError, "SDL has shut down.");

	if(!CD_INDRIVE(SDL_CDStatus(s_cd_ref)) || index >= s_cd_ref->numtracks
		|| index < 0)
	{
		return PyInt_FromLong(-1);
	}

	s_track_ref = &s_cd_ref->track[index];
																			 
	return Py_BuildValue("(iii)", s_track_ref->type, s_track_ref->length,
		s_track_ref->offset);	
}



static PyObject* cd_seq_get_slice(PyObject* self, int start, int end)
{
	PyCDObject* cd_ref = (PyCDObject*)self;
	SDL_CD* s_cd_ref = cd_ref->cd;
	PyObject* track_tuple;
	int count;
	int i;

	if(!SDL_WasInit(SDL_INIT_CDROM))
		return RAISE(PyExc_RuntimeError, "SDL has shut down.");
	
	i = min(start,end);
	end = max(start,end);
	start = i;

	if(start < 0)
		return RAISE(PyExc_IndexError, "Index out of range.");
	
	if(CD_INDRIVE(SDL_CDStatus(s_cd_ref)) <= 0)
		return RAISE(PyExc_RuntimeError, "CD unavailable");

	/* If end is greater than the real number of tracks, trunctate it.
	 * This is done because of how Python handles slicing
	*/
	end = min(end, s_cd_ref->numtracks);
	count = end - start;

	track_tuple = PyTuple_New(count);
	if(!track_tuple)
		return NULL;

	for(i = 0;i < count;i++)
	{
		PyObject* elem_tuple;
		SDL_CDtrack* s_track_ref = &s_cd_ref->track[i + start];

		elem_tuple = Py_BuildValue("(iii)",
			s_track_ref->type, s_track_ref->length, s_track_ref->offset);
		if(!elem_tuple)
		{
			Py_DECREF(track_tuple);
			return NULL;
		}

		PyTuple_SET_ITEM(track_tuple, i, elem_tuple);
	}
	return track_tuple; 	
}



static PyMethodDef cd_builtins[] =
{
	{ "play_tracks", cd_play_tracks, 1, cd_play_tracks_doc },
	{ "play", cd_play, 1, cd_play_doc },
	{ "pause", cd_pause, 1, cd_pause_doc },
	{ "resume", cd_resume, 1, cd_resume_doc },
	{ "stop", cd_stop, 1, cd_stop_doc },
	{ "eject", cd_eject, 1, cd_eject_doc },
	{ "get_status", cd_status, 1, cd_status_doc },
	{ "get_track", cd_cur_track, 1, cd_cur_track_doc },
	{ "get_frame", cd_cur_frame, 1, cd_cur_frame_doc },
	{ "get_empty", cd_get_empty, 1, cd_get_empty_doc },
	{ NULL, NULL }
};

static PyObject* cd_getattr(PyObject* self, char* attrname)
{
	if(SDL_WasInit(SDL_INIT_CDROM))
		return Py_FindMethod(cd_builtins, self, attrname);

	return RAISE(PyExc_NameError, attrname);
}



static PySequenceMethods cd_as_sequence = 
{
	cd_seq_len,
	0,
	0,
	cd_seq_get,
	cd_seq_get_slice,
	0,
	0
};

static PyTypeObject PyCD_Type =
{
	PyObject_HEAD_INIT(NULL)
	0,
	"CD",
	sizeof(PyCDObject),
	0,
	cd_dealloc,
	0,
	cd_getattr,
	0,
	0,
	0,
	0,
	&cd_as_sequence,
	0
};



static PyObject* PyCD_New(SDL_CD* cdrom)
{
	PyCDObject* cd;

	if(!cdrom)
		return RAISE(PyExc_SDLError, SDL_GetError());
	
	cd = PyObject_NEW(PyCDObject, &PyCD_Type);
	if(!cd) return NULL;

	cd->cd = cdrom;

	return (PyObject*)cd;
}





static PyMethodDef cdrom_builtins[] =
{
	{ "__PYGAMEinit__", cdrom_autoinit, 1, cdrom_init_doc },
	{ "init", cdrom_init, 1, cdrom_init_doc },
	{ "quit", cdrom_quit, 1, cdrom_quit_doc },
	{ "get_init", get_init, 1, get_init_doc },
	{ "get_count", cdrom_count, 1, cdrom_count_doc },
	{ "get_name", cdrom_name, 1, cdrom_name_doc },
	{ "open", cdrom_open, 1, cdrom_open_doc },
	{ NULL, NULL }
};



static char doc[] = "provides a low level interface to the C SDL API";

void initcdrom()
{
	PyObject *module, *dict, *apiobj;
	static void* c_api[PYGAMEAPI_CDROM_NUMSLOTS];

	PyType_Init(PyCD_Type);


    /* create the module */
	module = Py_InitModule3("cdrom", cdrom_builtins, doc);
	dict = PyModule_GetDict(module);

	/* export the c api */
	c_api[0] = &PyCD_Type;
	c_api[1] = PyCD_New;
	apiobj = PyCObject_FromVoidPtr(c_api, NULL);
	PyDict_SetItemString(dict, PYGAMEAPI_LOCAL_ENTRY, apiobj);

	/*imported needed apis*/
	import_pygame_base();
}

