/*
    PyGame - Python Game Library
    Copyright (C) 2000  Pete Shinners

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Pete Shinners
    pete@shinnerso.org
*/

#define PYGAMEAPI_JOYSTICK_INTERNAL
#include "pygame.h"




staticforward PyTypeObject Joystick_Type;
static PyObject* PyJoystick_New(SDL_Joystick*);
#define PyJoystick_Check(x) ((x)->ob_type == &PyCD_Type)



static void joy_autoquit()
{
	if(SDL_WasInit(SDL_INIT_JOYSTICK))
	{
		SDL_JoystickEventState(SDL_DISABLE);
		SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
	}
}

static PyObject* joy_autoinit(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	if(!SDL_WasInit(SDL_INIT_JOYSTICK))
	{
		if(SDL_InitSubSystem(SDL_INIT_JOYSTICK))
			return PyInt_FromLong(0);
		SDL_JoystickEventState(SDL_ENABLE);
		PyGame_RegisterQuit(cdrom_autoquit);
	}
	return PyInt_FromLong(1);
}


static char cdrom_quit_doc[] =
	"quit() -> None\n"
	"uninitialize the joystick subsystem";

static PyObject* joy_quit(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	joy_autoquit();

	RETURN_NONE
}


static char doc_joy_init_doc[] =
	"init() -> None\n"
	"initialize the joystick subsystem";

static PyObject* joy_init(PyObject* self, PyObject* arg)
{
	PyObject* result;
	int istrue;

	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	result = joy_autoinit(self, arg);
	istrue = PyObject_IsTrue(result);
	Py_DECREF(result);
	if(!istrue)
		return RAISE(PyExc_SDLError, SDL_GetError());

	RETURN_NONE
}



static char doc_get_init[] = 
	"get_init() -> bool\n"
	"returns true if joystick subsytem is initialized";

static PyObject* get_init(PyObject* self, PyObject* arg)
{
	if(!PyArg_ParseTuple(arg, ""))
		return NULL;

	return PyInt_FromLong(SDL_WasInit(SDL_INIT_JOYSTICK)!=0);
}



/*joystick object funcs*/


static void joy_dealloc(PyObject* self)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(SDL_WasInit(SDL_INIT_JOYSTICK)
		SDL_JoystickClose(joy_ref->joy);

	PyMem_DEL(self);
}


static char doc_joy_get_id[] =
	"get_id() -> id\n"
	"Returns the id of the joystick object.";

static PyObject* joy_get_id(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_JoystickIndex(joy_ref->joy));
}



static char doc_joy_get_axes[] =
	"get_axes() -> count\n"
	"Returns the number of axis controls.";

static PyObject* joy_get_axes(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_JoystickNumAxes(joy_ref->joy));
}



static char doc_joy_get_balls[] =
	"get_balls() -> count\n"
	"Returns the number of trackballs.";

static PyObject* joy_get_balls(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_JoystickNumBalls(joy_ref->joy));
}



static char doc_joy_get_hats[] =
	"get_hats() -> count\n"
	"Returns the number of POV hats.";

static PyObject* joy_get_hats(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_JoystickNumHats(joy_ref->joy));
}



static char doc_joy_get_buttons[] =
	"get_buttons() -> count\n"
	"Returns the number of buttons.";

static PyObject* joy_get_buttons(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;

	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyInt_FromLong(SDL_JoystickNumButtons(joy_ref->joy));
}



static char doc_joy_get_axis[] =
	"get_axis(axis) -> position\n"
	"Returns the current position of an axis control on the joystick.\n"
	"The values range from -1.0 to 1.0.";

static PyObject* joy_get_axis(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;
	int axis;

	if(!PyArg_ParseTuple(args, "i", &axis))
		return NULL;

	return PyFloat_FromDouble(SDL_JoystickGetAxis(joy_ref->joy, axis)/32767.0);
}



static char doc_joy_get_hat[] =
	"get_hat(pov_hat) -> state\n"
	"Returns the current state of a POV hat on the joystick.\n"
	"HAT_LEFTUP   HAT_UP       HAT_RIGHTUP\n"
	"HAT_LEFT     HAT_CENTERED HAT_RIGHT \n"
	"HAT_LEFTDOWN HAT_DOWN     HAT_RIGHTDOWN\n"; 

static PyObject* joy_get_hat(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;
	int hat;

	if(!PyArg_ParseTuple(args, "i", &hat))
		return NULL;

	return PyInt_FromLong(SDL_JoystickGetHat(joy_ref->joy, hat));
}



static char doc_joy_get_button[] =
	"get_button(button) -> int\n"
	"Indicates whether the joystick button is down or not.";

static PyObject* joy_get_button(PyObject* self, PyObject* args)
{
	PyJoystickObject* joy_ref = (PyJoystickObject*)self;
	int button;

	if(!PyArg_ParseTuple(args, "i", &button))
		return NULL;

	return PyInt_FromLong(SDL_JoystickGetButton(joy_ref->joy, button));
}



/*joystick module funcs*/

static PyObject* PyJoystick_New(SDL_Joystick* joy)
{
	PyJoystickObject* joyobj;

	if(!joy)
		return RAISE(PyExc_SDLError, SDL_GetError());

	joyobj = PyObject_NEW(PyJoystickObject, &Joystick_Type);
	if(!joyobj)
		return NULL;

	joyobj->joy = joy;

	return (PyObject*)joyobj;
}



static char doc_joy_open[] =
	"open(id) -> Joystick\n"
	"Opens the specified joystick, and returns an object for manipulating it.";

static PyObject* joy_open(PyObject* self, PyObject* args)
{
	int id;

	if(!PyArg_ParseTuple(args, "i", &id))
		return NULL;

	JOY_INIT_CHECK

	return PyJoystick_New(SDL_JoystickOpen(id));
}



static char doc_joy_get_count[] =
	"get_count() -> int\n"
	"Returns the number of joysticks attached to the system.";

static PyObject* joy_get_count(PyObject* self, PyObject* args)
{
	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	JOY_INIT_CHECK

	return PyInt_FromLong(SDL_NumJoysticks());
}



static char doc_joy_get_name[] =
	"get_name(id) -> string\n"
	"Returns the name of the joystick identified by id.";

static PyObject* joy_get_name(PyObject* self, PyObject* args)
{
	int id;

	if(!PyArg_ParseTuple(args, "i", &id))
		return NULL;

	JOY_INIT_CHECK

	return PyString_FromString(SDL_JoystickName(id));
}



static char doc_joy_is_opened[] =
	"is_opened(id) -> int\n"
	"Returns whether the joystick identified by id has been opened or not.";

static PyObject* joy_is_opened(PyObject* self, PyObject* args)
{
	int id;
	
	if(!PyArg_ParseTuple(args, "i", &id))
		return NULL;

	JOY_INIT_CHECK

	return PyInt_FromLong(SDL_JoystickOpened(id));
}




static PyMethodDef joy__builtins__[] =
{
	{ "get_id", joy_get_id, 1, joy_get_id },
	{ "get_axes", joy_get_axes, 1, joy_get_axes },
	{ "get_balls", joy_get_balls, 1, joy_get_balls },
	{ "get_hats", joy_get_hats, 1, joy_get_hats },
	{ "get_buttons", joy_get_buttons, 1, joy_get_buttons },
	{ "get_axis", joy_get_axis, 1, joy_get_axis },
	{ "get_hat", joy_get_hat, 1, joy_get_hat },
	{ "get_button", joy_get_button, 1, joy_get_button },
	{ NULL, NULL }
};

static PyObject* joy_getattr(PyObject* self, char* attrname)
{
	if(_pysdl_sdl_is_initialized && _pysdl_joy_is_initialized)
		return Py_FindMethod(joy__builtins__, self, attrname);

	PyErr_SetString(PyExc_NameError,	attrname);
	return NULL;
}



static PyTypeObject Joystick_Type =
{
	PyObject_HEAD_INIT(NULL)
	0,
	"Joystick",
	sizeof(PyJoystickObject),
	0,
	joy_dealloc,
	0,
	joy_getattr
};


static PyMethodDef joystick_builtins[] =
{
	{ "__PYGAMEinit__", joy_autoinit, 1, joy_init_doc },
	{ "init", joy_init, 1, joy_init_doc },
	{ "quit", joy_quit, 1, joy_quit_doc },
	{ "get_count", joy_get_count, 1, joy_get_count },
	{ "get_name", joy_get_name, 1, joy_get_name },
	{ "open", joy_open, 1, joy_open },
	{ "is_opened", joy_is_opened, 1, joy_is_opened },
	{ "update", joy_update, 1, joy_update },
	{ "event_state", joy_event_state, 1, joy_event_state },
	{ NULL, NULL }
};


static char doc[] = "provides a low level interface to the C SDL API";

void initcdrom()
{
	PyObject *module, *dict, *apiobj;
	static void* c_api[PYGAMEAPI_CDROM_NUMSLOTS];

	PyType_Init(PyCD_Type);


    /* create the module */
	module = Py_InitModule3("joystick", joystick_builtins, doc);
	dict = PyModule_GetDict(module);

	/* export the c api */
	c_api[0] = &PyJoystick_Type;
	c_api[1] = PyJoystick_New;
	apiobj = PyCObject_FromVoidPtr(c_api, NULL);
	PyDict_SetItemString(dict, PYGAMEAPI_LOCAL_ENTRY, apiobj);

	/*imported needed apis*/
	import_pygame_base();
}



