/*
    PyGame - Python Game Library
    Copyright (C) 2000  Pete Shinners

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Pete Shinners
    pete@shinnerso.org
*/

/*
 *  Python Rect Object -- useful 2d rectangle class
 */
#define PYGAMEAPI_RECT_INTERNAL
#include "pygame.h"


staticforward PyTypeObject PyRect_Type;
#define PyRect_Check(x) ((x)->ob_type == &PyRect_Type)



GAME_Rect* GameRect_FromObject(PyObject* obj, GAME_Rect* temp)
{
	short val;

	if(PyRect_Check(obj))
		return &((PyRectObject*)obj)->r;
	if(PySequence_Check(obj))
	{
		if(PySequence_Length(obj) == 4)
		{
			if(!ShortFromObjIndex(obj, 0, &val)) return NULL; temp->x = val;
			if(!ShortFromObjIndex(obj, 1, &val)) return NULL; temp->y = val;
			if(!ShortFromObjIndex(obj, 2, &val)) return NULL; temp->w = val;
			if(!ShortFromObjIndex(obj, 3, &val)) return NULL; temp->h = val;
			return temp;
		}
		if(PySequence_Length(obj) == 2)
		{
			PyObject* sub = PySequence_GetItem(obj, 0);
			if(!sub || !PySequence_Check(sub)) {Py_XDECREF(sub); return NULL;}
			if(!ShortFromObjIndex(sub, 0, &val)) {Py_DECREF(sub); return NULL;} temp->x = val;
			if(!ShortFromObjIndex(sub, 1, &val)) {Py_DECREF(sub); return NULL;} temp->y = val;
			sub = PySequence_GetItem(obj, 1);
			if(!sub || !PySequence_Check(sub)) {Py_XDECREF(sub); return NULL;}
			if(!ShortFromObjIndex(sub, 0, &val)) {Py_DECREF(sub); return NULL;} temp->w = val;
			if(!ShortFromObjIndex(sub, 1, &val)) {Py_DECREF(sub); return NULL;} temp->h = val;
			return temp;
		}
		if(PyTuple_Check(obj) && PyTuple_Size(obj) == 1) /*looks like an arg?*/
		{
			PyObject* sub = PyTuple_GET_ITEM(obj, 0);
			if(sub)
				return GameRect_FromObject(sub, temp);
		}
	}
	return NULL;
}

static int Rect_SetTop(GAME_Rect* r, short val)
{
	r->y = val;
	return 0;
}	
static int Rect_SetBottom(GAME_Rect* r, short val)
{
	r->y = val - r->h;
	return 0;
}	
static int Rect_SetLeft(GAME_Rect* r, short val)
{
	r->x = val;
	return 0;
}	
static int Rect_SetRight(GAME_Rect* r, short val)
{
	r->x = val - r->w;
	return 0;
}
static int Rect_SetWidth(GAME_Rect* r, short val)
{
	r->x -= val - r->w;
	r->w = val;
	return 0;
}
static int Rect_SetHeight(GAME_Rect* r, short val)
{
	r->y -= val - r->h;
	r->h = val;
	return 0;
}

PyObject* PyRect_New(GAME_Rect* r)
{
	PyRectObject* rect = PyObject_NEW(PyRectObject, &PyRect_Type);
	if(!rect)
		return NULL;
	
	rect->r.x = r->x;
	rect->r.y = r->y;
	rect->r.w = r->w;
	rect->r.h = r->h;

	return (PyObject*)rect;
}

PyObject* PyRect_New4(short x, short y, short w, short h)
{
	PyRectObject* rect = PyObject_NEW(PyRectObject, &PyRect_Type);
	if(!rect)
		return NULL;
	
	rect->r.x = x;
	rect->r.y = y;
	rect->r.w = w;
	rect->r.h = h;

	return (PyObject*)rect;
}

static int DoRectsIntersect(GAME_Rect *A, GAME_Rect *B)
{
	if(!((A->x >= B->x) && (A->x < (B->x+B->w))) &&
	   !((B->x >= A->x) && (B->x < (A->x+A->w))) &&
	   !(((A->x+A->w) > B->x) && ((A->x+A->w) <= (B->x+B->w))) &&
	   !(((B->x+B->w) > A->x) && ((B->x+B->w) <= (A->x+A->w))))
		return 0;
	if(!((A->y >= B->y) && (A->y < (B->y+B->h))) &&
	   !((B->y >= A->y) && (B->y < (A->y+A->h))) &&
	   !(((A->y+A->h) > B->y) && ((A->y+A->h) <= (B->y+B->h))) &&
	   !(((B->y+B->h) > A->y) && ((B->y+B->h) <= (A->y+A->h))))
		return 0;

	return 1;
}




/* main rect methods */

static PyObject* rect_normalize(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	if(!PyArg_ParseTuple(args, ""))
		return NULL;

	if((short)self->r.w < 0)
	{
		self->r.x += self->r.w;
		self->r.w = -self->r.w;
	}
	if((short)self->r.h < 0)
	{
		self->r.y += self->r.h;
		self->r.h = -self->r.h;
	}

	RETURN_NONE
}
static char doc_normalize[] =
	"rect.normalize() -> None\n"
	"Fix negative width/height for this rect";


static PyObject* rect_move(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	short x, y;

	if(!TwoShortsFromObj(args, &x, &y))
		return RAISE(PyExc_TypeError, "argument must contain two numbers");

	return PyRect_New4((short)(self->r.x+x), (short)(self->r.y+y), self->r.w, self->r.h);
}
static char doc_move[] =
	"rect.move(x, y) -> rect\n"
	"Returns a rect moved by the given offset";


static PyObject* rect_inflate(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	short x, y;

	if(!TwoShortsFromObj(args, &x, &y))
		return RAISE(PyExc_TypeError, "argument must contain two numbers");

	return PyRect_New4((short)(self->r.x-x/2), (short)(self->r.y-y/2), (short)(self->r.w+x), (short)(self->r.h+y));
}
static char doc_inflate[] =
	"rect.inflate(x, y) -> rect\n"
	"Returns a rect with the dimensions changed by the arguments (from center).";


static PyObject* rect_union(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	short x, y, w, h;
	if(!(argrect = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	x = min(self->r.x, argrect->x);
	y = min(self->r.y, argrect->y);
	w = max(self->r.x+self->r.w, argrect->x+argrect->w) - x;
	h = max(self->r.y+self->r.h, argrect->y+argrect->h) - y;
	return PyRect_New4(x, y, w, h);
}
static char doc_union[] =
	"rect.union(rectstyle) -> rect\n"
	"Returns a rectangle surrounding both rectangles.";


static PyObject* rect_collidepoint(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	short x, y;
	int inside;

	if(!TwoShortsFromObj(args, &x, &y))
		return RAISE(PyExc_TypeError, "argument must contain two numbers");

	inside = x>=self->r.x && x<=self->r.x+self->r.w && 
				y>=self->r.y && y<=self->r.y+self->r.h;

	return PyInt_FromLong(inside);
}
static char doc_collidepoint[] =
	"rect.collidepoint(x, y) -> int\n"
	"Returns true if the given position is inside the rect (inclusive)";



static PyObject* rect_colliderect(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	if(!(argrect = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	return PyInt_FromLong(DoRectsIntersect(&self->r, argrect));
}
static char doc_colliderect[] =
	"rect.colliderect(rectstyle) -> rect\n"
	"Returns true if rectangles overlap.";


static PyObject* rect_collidelist(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	int loop, size;
	PyObject* list, *obj;
	PyObject* ret = NULL;

	if(!PyArg_ParseTuple(args, "O", &list))
		return NULL;

	if(!PySequence_Check(list))
		return RAISE(PyExc_TypeError, "Argument must be a sequence of rectstyle objects.");

	size = PySequence_Length(list); /*warning, size could be -1 on error?*/
	for(loop = 0; loop < size; ++loop)
	{
		obj = PySequence_GetItem(list, loop);
		if(!obj || !(argrect = GameRect_FromObject(obj, &temp)))
		{	
			RAISE(PyExc_TypeError, "Argument must be a sequence of rectstyle objects.");
			Py_XDECREF(obj);
			break;
		}
		if(DoRectsIntersect(&self->r, argrect))
		{
			ret = PyInt_FromLong(loop);
			Py_DECREF(obj);
			break;
		}
		Py_DECREF(obj);
	}
	if(loop == size)
		ret = PyInt_FromLong(-1);

	return ret;	
}
static char doc_collidelist[] =
	"rect.colliderect(rectstyle list) -> int index\n"
	"Returns first index of rectangle that intersects from a list. -1 if no intersections.";


static PyObject* rect_collidelistall(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	int loop, size;
	PyObject* list, *obj;
	PyObject* ret = NULL;

	if(!PyArg_ParseTuple(args, "O", &list))
		return NULL;

	if(!PySequence_Check(list))
		return RAISE(PyExc_TypeError, "Argument must be a sequence of rectstyle objects.");

	ret = PyList_New(0);
	if(!ret)
		return NULL;

	size = PySequence_Length(list); /*warning, size could be -1?*/
	for(loop = 0; loop < size; ++loop)
	{
		obj = PySequence_GetItem(list, loop);

		if(!obj || !(argrect = GameRect_FromObject(obj, &temp)))
		{	
			Py_XDECREF(obj);
			Py_DECREF(ret);
			return RAISE(PyExc_TypeError, "Argument must be a sequence of rectstyle objects.");
		}

		if(DoRectsIntersect(&self->r, argrect))
		{
			PyObject* num = PyInt_FromLong(loop);
			if(!num)
			{
				Py_DECREF(obj);
				return NULL;
			}
			PyList_Append(ret, num);
			Py_DECREF(num);
		}
		Py_DECREF(obj);
	}

	return ret;	
}
static char doc_collidelistall[] =
	"rect.colliderectall(rectstyle list) -> int index\n"
	"Returns list of all indexes of rects intersecting.";


static PyObject* rect_clip(PyObject* self, PyObject* args)
{
	GAME_Rect *A, *B, temp;
	short x, y, w, h;

	A = &((PyRectObject*)self)->r;
	if(!(B = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	/* Left */
	if((A->x >= B->x) && (A->x < (B->x+B->w)))
		x = A->x;
	else if((B->x >= A->x) && (B->x < (A->x+A->w)))
		x = B->x;
	else
		goto nointersect;
	/* Right */
	if(((A->x+A->w) > B->x) && ((A->x+A->w) <= (B->x+B->w)))
		w = (A->x+A->w) - x;
	else if(((B->x+B->w) > A->x) && ((B->x+B->w) <= (A->x+A->w)))
		w = (B->x+B->w) - x;
	else
		goto nointersect;
	
	/* Top */
	if((A->y >= B->y) && (A->y < (B->y+B->h)))
		y = A->y;
	else if((B->y >= A->y) && (B->y < (A->y+A->h)))
		y = B->y;
	else
		goto nointersect;
	/* Bottom */
	if (((A->y+A->h) > B->y) && ((A->y+A->h) <= (B->y+B->h)))
		h = (A->y+A->h) - y;
	else if(((B->y+B->h) > A->y) && ((B->y+B->h) <= (A->y+A->h)))
		h = (B->y+B->h) - y;
	else
		goto nointersect;

	return PyRect_New4(x, y, w, h);

nointersect:
	return PyRect_New4(A->x, A->y, 0, 0);
}
static char doc_clip[] =
	"rect.clip(rectstyle) -> rect\n"
	"Returns a rect of the overlapping area between rects. 0 width/height if not touching.";


static PyObject* rect_contains(PyObject* oself, PyObject* args)
{
	int contained;
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	if(!(argrect = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	contained = (self->r.x <= argrect->x) && (self->r.y <= argrect->y) &&
					(self->r.x + self->r.w >= argrect->x + argrect->w) &&
					(self->r.y + self->r.h >= argrect->y + argrect->h);

	return PyInt_FromLong(contained);
}
static char doc_contains[] =
	"rect.contains(rectstyle) -> int\n"
	"Returns true if rect completely surrounds given rect.";


static PyObject* rect_clamp(PyObject* oself, PyObject* args)
{
	PyRectObject* self = (PyRectObject*)oself;
	GAME_Rect *argrect, temp;
	short x, y;
	if(!(argrect = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	if(self->r.w >= argrect->w)
		x = (argrect->x+argrect->w) / 2 - self->r.w / 2;
	else if(self->r.x < argrect->x)
		x = argrect->x;
	else if(self->r.x + self->r.w > argrect->x + argrect->w)
		x = argrect->x + argrect->w - self->r.w;
	else
		x = self->r.x;

	if(self->r.h >= argrect->h)
		y = (argrect->y+argrect->h) / 2 - self->r.h / 2;
	else if(self->r.y < argrect->y)
		y = argrect->y;
	else if(self->r.y + self->r.h > argrect->y + argrect->h)
		y = argrect->y + argrect->h - self->r.h;
	else
		y = self->r.y;

	return PyRect_New4(x, y, self->r.w, self->r.h);
}
static char doc_clamp[] =
	"rect.clamp(rectstyle) -> rect\n"
	"Returns rect moved to be contained in given rectangle.";


static struct PyMethodDef rect_methods[] =
{
	{"normalize",		(PyCFunction)rect_normalize,	1, doc_normalize},
	{"move",			(PyCFunction)rect_move,			1, doc_move},
	{"inflate",			(PyCFunction)rect_inflate,		1, doc_inflate},		
	{"clip",			(PyCFunction)rect_clip,			1, doc_clip},
	{"union",			(PyCFunction)rect_union,		1, doc_union},
	{"collidepoint",	(PyCFunction)rect_collidepoint,	1, doc_collidepoint},
	{"colliderect",		(PyCFunction)rect_colliderect,	1, doc_colliderect},
	{"collidelist",		(PyCFunction)rect_collidelist,	1, doc_collidelist},
	{"collidelistall",	(PyCFunction)rect_collidelistall,1,doc_collidelistall},
	{"contains",		(PyCFunction)rect_contains,		1, doc_contains},
	{"clamp",			(PyCFunction)rect_clamp,		1, doc_clamp},
/* these are totally unwritten. volunteers? */
/*	{"cleanup",			(PyCFunction)rect_cleanup,		1, doc_cleanup}, */
/*	{"remove",			(PyCFunction)rect_remove,		1, doc_remove}, */
	{NULL,		NULL}
};




/* sequence functions */

static int rect_length(PyRectObject *self)
{
	return 4;
}

static PyObject* rect_item(PyRectObject *self, int i)
{
	short* data = (short*)&self->r;
	if(i<0 || i>3)
		return RAISE(PyExc_IndexError, "Invalid rect Index");

	return PyInt_FromLong(data[i]);
}

static int rect_ass_item(PyRectObject *self, int i, PyObject *v)
{
	short val;
	short* data = (short*)&self->r;
	if(i<0 || i>3)
	{
		RAISE(PyExc_IndexError, "Invalid rect Index");
		return -1;
	}
	if(!ShortFromObj(v, &val))
	{
		RAISE(PyExc_TypeError, "Must assign numeric values");
		return -1;
	}
	data[i] = (short)val;
	return 0;
}


static PyObject* rect_slice(PyRectObject *self, int ilow, int ihigh)
{
	PyObject *list;
	short* data = (short*)&self->r;
	int numitems, loop, l = 4;
	
	if (ihigh < 0) ihigh += l;
	if (ilow  < 0) ilow  += l;
	if (ilow < 0) ilow = 0;
	else if (ilow > l) ilow = l;
	if (ihigh < 0) ihigh = 0;
	else if (ihigh > l) ihigh = l;
	if (ihigh < ilow) ihigh = ilow;

	numitems = ihigh - ilow;
	list = PyList_New(numitems);
	for(loop = 0; loop < numitems; ++loop)
		PyList_SET_ITEM(list, loop, PyInt_FromLong(data[loop+ilow]));

	return list;
}



static int rect_ass_slice(PyRectObject *self, int ilow, int ihigh, PyObject *v)
{
	short* data = (short*)&self->r;
	int numitems, loop, l = 4;
	short val;
	
	if(!PySequence_Check(v))
	{
		RAISE(PyExc_TypeError, "Assigned slice must be a sequence");
		return -1;
	}

	if (ihigh < 0) ihigh += l;
	if (ilow  < 0) ilow  += l;
	if (ilow < 0) ilow = 0;
	else if (ilow > l) ilow = l;
	if (ihigh < 0) ihigh = 0;
	else if (ihigh > l) ihigh = l;
	if (ihigh < ilow) ihigh = ilow;

	numitems = ihigh - ilow;
	if(numitems != PySequence_Length(v))
	{
		RAISE(PyExc_ValueError, "Assigned slice must be same length");
		return -1;
	}

	for(loop = 0; loop < numitems; ++loop)
	{
		if(!ShortFromObjIndex(v, loop, &val)) return -1;
		data[loop+ilow] = val;
	}

	return 0;
}

static PySequenceMethods rect_as_sequence = {
	(inquiry)rect_length,				/*length*/
	(binaryfunc)NULL,					/*concat*/
	(intargfunc)NULL,					/*repeat*/
	(intargfunc)rect_item,				/*item*/
	(intintargfunc)rect_slice,			/*slice*/
	(intobjargproc)rect_ass_item,		/*ass_item*/
	(intintobjargproc)rect_ass_slice,	/*ass_slice*/
};



/* numeric functions */

static int rect_nonzero(PyRectObject *self)
{
	return self->r.w != 0 && self->r.h != 0;
}

static int rect_coerce(PyObject** o1, PyObject** o2)
{
	PyObject* new1;
	PyObject* new2;
	GAME_Rect* r, temp;

	if(PyRect_Check(*o1))
	{
		new1 = *o1;
		Py_INCREF(new1);
	}
	else if(r = GameRect_FromObject(*o1, &temp))
		new1 = PyRect_New(r);
	else
		return 1;

	if(PyRect_Check(*o2))
	{
		new2 = *o2;
		Py_INCREF(new2);
	}
	else if(r = GameRect_FromObject(*o2, &temp))
		new2 = PyRect_New(r);
	else
	{
		Py_DECREF(new1);
		return 1;
	}

	*o1 = new1;
	*o2 = new2;
	return 0;
}

static PyNumberMethods rect_as_number = {
	(binaryfunc)NULL,		/*add*/
	(binaryfunc)NULL,		/*subtract*/
	(binaryfunc)NULL,		/*multiply*/
	(binaryfunc)NULL,		/*divide*/
	(binaryfunc)NULL,		/*remainder*/
	(binaryfunc)NULL,		/*divmod*/
	(ternaryfunc)NULL,		/*power*/
	(unaryfunc)NULL,		/*negative*/
	(unaryfunc)NULL,		/*pos*/ 
	(unaryfunc)NULL,		/*abs*/
	(inquiry)rect_nonzero,	/*nonzero*/
	(unaryfunc)NULL,		/*invert*/
	(binaryfunc)NULL,		/*lshift*/
	(binaryfunc)NULL,		/*rshift*/
	(binaryfunc)NULL,		/*and*/
	(binaryfunc)NULL,		/*xor*/
	(binaryfunc)NULL,		/*or*/
	(coercion)rect_coerce,	/*coerce*/
	(unaryfunc)NULL,		/*int*/
	(unaryfunc)NULL,		/*long*/
	(unaryfunc)NULL,		/*float*/
	(unaryfunc)NULL,		/*oct*/
	(unaryfunc)NULL,		/*hex*/
};


/* object type functions */
static void rect_dealloc(PyRectObject *self)
{
	PyMem_DEL(self);
}


static PyObject *rect_repr(PyRectObject *self)
{
	char string[256];	
	sprintf(string, "<rect(%d, %d, %d, %d)>", self->r.x, self->r.y, self->r.w, self->r.h);	
	return PyString_FromString(string);
}


static PyObject *rect_str(PyRectObject *self)
{
	return rect_repr(self);
}


static int rect_compare(PyRectObject *self, PyObject *other)
{
	GAME_Rect *orect, temp;

	orect = GameRect_FromObject(other, &temp);
	if(!orect)
	{
		RAISE(PyExc_TypeError, "must compare rect with rect style object");
		return -1;
	}

	if(self->r.x != orect->x)
		return self->r.x < orect->x ? -1 : 1;
	if(self->r.y != orect->y)
		return self->r.y < orect->y ? -1 : 1;
	if(self->r.x != orect->x)
		return self->r.w < orect->w ? -1 : 1;
	if(self->r.x != orect->x)
		return self->r.h < orect->h ? -1 : 1;

	return 0;
}


static PyObject *rect_getattr(PyRectObject *self, char *name)
{
	PyObject *ret = NULL;
	GAME_Rect *r = &self->r;
	
	if(!strcmp(name, "top"))
		ret = PyInt_FromLong(r->y);
	else if(!strcmp(name, "bottom"))
		ret = PyInt_FromLong(r->y + r->h);
	else if(!strcmp(name, "left"))
		ret = PyInt_FromLong(r->x);
	else if(!strcmp(name, "right"))
		ret = PyInt_FromLong(r->x + r->w);
	else if(!strcmp(name, "width"))
		ret = PyInt_FromLong(r->w);
	else if(!strcmp(name, "height"))
		ret = PyInt_FromLong(r->h);
	else if(!strcmp(name, "centerx"))
		ret = PyInt_FromLong(r->x+r->w/2);
	else if(!strcmp(name, "centery"))
		ret = PyInt_FromLong(r->y+r->h/2);
	else if(!strcmp(name, "topleft"))
		ret = Py_BuildValue("(ii)", r->x, r->y);
	else if(!strcmp(name, "bottomleft"))
		ret = Py_BuildValue("(ii)", r->x, r->y + r->h);
	else if(!strcmp(name, "topright"))
		ret = Py_BuildValue("(ii)", r->x + r->w, r->y);
	else if(!strcmp(name, "bottomright"))
		ret = Py_BuildValue("(ii)", r->x + r->w, r->y + r->h);
	else if(!strcmp(name, "size"))
		ret = Py_BuildValue("(ii)", r->w, r->h);
	else if(!strcmp(name, "center"))
		ret = Py_BuildValue("(ii)", r->x + r->w / 2, r->y + r->h / 2);
	else
		ret = Py_FindMethod(rect_methods, (PyObject *)self, name);

	return ret;
}

static int rect_setattr(PyRectObject *self, char *name, PyObject *op)
{
	int ret = -1;
	short val1, val2;
	GAME_Rect *r = &self->r;
	
	if(!strcmp(name, "top"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetTop(r, val1);
	}	
	else if(!strcmp(name, "bottom"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetBottom(r, val1);
	}	
	else if(!strcmp(name, "left"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetLeft(r, val1);
	}	
	else if(!strcmp(name, "right"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetRight(r, val1);
	}	
	else if(!strcmp(name, "width"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetWidth(r, val1);
	}	
	else if(!strcmp(name, "height"))
	{
		if(ShortFromObj(op, &val1))
			ret = Rect_SetHeight(r, val1);
	}	
	else if(!strcmp(name, "topleft"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			Rect_SetLeft(r, val1);
			ret = Rect_SetTop(r, val2);
		}
	}	
	else if(!strcmp(name, "bottomleft"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			Rect_SetLeft(r, val1);
			ret = Rect_SetBottom(r, val2);
		}
	}	
	else if(!strcmp(name, "topright"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			Rect_SetRight(r, val1);
			ret = Rect_SetTop(r, val2);
		}
	}	
	else if(!strcmp(name, "bottomright"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			Rect_SetRight(r, val1);
			ret = Rect_SetBottom(r, val2);
		}
	}	
	else if(!strcmp(name, "size"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			Rect_SetWidth(r, val1);
			ret = Rect_SetHeight(r, val2);
		}
	}	
	else if(!strcmp(name, "center"))
	{
		if(TwoShortsFromObj(op, &val1, &val2))
		{
			r->x += val1 - (r->x + r->w / 2);
			r->y += val2 - (r->y + r->h / 2);
			ret = 0;
		}
	}
	else if(!strcmp(name, "centerx"))
	{
		if(ShortFromObj(op, &val1))
		{
			r->x += val1 - (r->x + r->w / 2);
			ret = 0;
		}
	}
	else if(!strcmp(name, "centery"))
	{
		if(ShortFromObj(op, &val1))
		{
			r->y += val1 - (r->y + r->h / 2);
			ret = 0;
		}
	}
	else
	{
		RAISE(PyExc_AttributeError, "Attribute cannot be modified");
		return -1;
	}

	if(ret == -1)
		RAISE(PyExc_TypeError, "Unable to assign to rect attribute");

	return ret;
}

static char Recttype__doc__[] = 
"SDL_Rect style rectangle, enhanced";


static PyTypeObject PyRect_Type = { 
	PyObject_HEAD_INIT(0)
	0,							/*size*/
	"rect",						/*name*/
	sizeof(PyRectObject),		/*basicsize*/
	0,							/*itemsize*/
	/* methods */
	(destructor)rect_dealloc,	/*dealloc*/
	(printfunc)NULL,			/*print*/
	(getattrfunc)rect_getattr,	/*getattr*/
	(setattrfunc)rect_setattr,	/*setattr*/
	(cmpfunc)rect_compare,		/*compare*/
	(reprfunc)rect_repr,		/*repr*/
	&rect_as_number,			/*as_number*/
	&rect_as_sequence,			/*as_sequence*/
	NULL,			 			/*as_mapping*/
	(hashfunc)NULL,				/*hash*/
	(ternaryfunc)NULL,			/*call*/
	(reprfunc)rect_str,			/*str*/

	/* Space for future expansion */
	0L,0L,0L,0L,
	Recttype__doc__ /* Documentation string */
};



/*module globals*/
static PyObject* rect_new(PyObject* self, PyObject* args)
{
	GAME_Rect *argrect, temp;
	if(!(argrect = GameRect_FromObject(args, &temp)))
		return RAISE(PyExc_TypeError, "Argument must be rect style object");

	return PyRect_New4(argrect->x, argrect->y, argrect->w, argrect->h);
}

static char rect_new__doc[] =
	"creates a new rectangle object";


static PyMethodDef rect__builtins__[] =
{
	{ "rect", rect_new, 1, rect_new__doc }, 
	{NULL, NULL}
};




void initrect()
{
	PyObject *module, *dict, *apiobj;
	static void* c_api[PYGAMEAPI_RECT_NUMSLOTS];

	/* Create the module and add the functions */
	PyType_Init(PyRect_Type);


	module = Py_InitModule3("rect", rect__builtins__, "rectangle object");
	dict = PyModule_GetDict(module);

	/* export the c api */
	c_api[0] = &PyRect_Type;
	c_api[1] = PyRect_New;
	c_api[2] = PyRect_New4;
	c_api[3] = GameRect_FromObject;
	apiobj = PyCObject_FromVoidPtr(c_api, NULL);
	PyDict_SetItemString(dict, PYGAMEAPI_LOCAL_ENTRY, apiobj);

	/*imported needed apis*/
	import_pygame_base();
}
